#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <sys/time.h>
#include <assert.h>

#include <cblas.h>
#include <lapacke.h>

// array_rand initializes an array with random values
void array_rand(double* array, int size) {
  for(int i=0; i<size; i++) {
    array[i] = (double)rand() / (double)(RAND_MAX - 1); // random number in [0, 1[;
  }
}

// Compute b:= A*xref and then solve A*x=b.
int main() {
  srand(421);

  int n = 1000;

  //
  // Initialization: Generating a random matrix A and a random vector xref.
  //


  double* A    = (double *)malloc(n*n*sizeof(double));

  double* xref = (double *)malloc(n*sizeof(double)); // reference solution
  double* b    = (double *)malloc(n*sizeof(double));

  if (A == NULL || xref == NULL || b == NULL) {
    perror("malloc failed");
    free(A); free(xref); free(b);
    return EXIT_FAILURE;
  }

  array_rand(A, n*n);
  array_rand(xref, n);

  //
  // Perform b := A*xref
  //

  // DGEMV (y := alpha*A'*x + beta*y)
  cblas_dgemv(CblasRowMajor, CblasNoTrans, 
	      n, n,        /* number of rows and columns of A */
	      1.0, A, n,   /* alpha, A, lda */
	      xref, 1,     /* x, incx */
	      0.0, b, 1);  /* beta, y, incy */

  //
  // Solve A*x = b
  //

  // LU decomposition with partial pivoting (A = P*L*U)
  int* P = (int *)malloc(n*sizeof(int)); // permutation matrix
  if (P == NULL) {
    perror("malloc failed");
    free(A); free(xref); free(b);
    return EXIT_FAILURE;
  }

  int info = LAPACKE_dgesv(LAPACK_ROW_MAJOR, 
			   n,     /* order of the matrix A */
			   1,     /* number of right hand sides */
			   A, n,  /* A, lda */
			   P,     /* ipiv */
			   b, 1); /* b on entry/x on exit, ldb */
  free(P);

  if (info < 0)      fprintf(stderr, "error: the %i-th argument had an illegal value.\n", -info);
  else if (info > 0) fprintf(stderr, "error: U is singular. U(%d,%d) is exactly zero.\n", info, info);
  else printf("DGESV: OK\n");

  //
  // Finalize
  //

  // Test if b == xref
  for (int i = 0; i<n; i++) { assert(fabs(b[i] - xref[i])<1e-10); }

  free(A);
  free(b);
  free(xref);

  return EXIT_SUCCESS;
}
